#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include <ESPmDNS.h>
#include <Wire.h>
#include "wifi/wifi.h"
#include "esp_http_server.h"
#include <EEPROM.h>

const	char	firm_ver[ 16] = { "1.0"};

#define EEP_SCL		19
#define EEP_SDA		18

#define BMS_ADR		0x18
#define INA226_ADR	0x40

#define LEDG		19
#define LEDR		18

#define	RLY_ON		5
#define	RLY_SEL1	10
#define	RLY_SEL2	7
#define	RLY_SEL3	6

#define	I2C_SDA		1
#define	I2C_SCL		2

#define	uchar	unsigned char
#define	ushort	unsigned short
#define	ulong	unsigned long
#define	uint	unsigned int

const IPAddress 	subnet(255, 255, 255, 0);

#define	EEP_SIZE	1024
int	charge_seq;

int	max_cell;
int	min_cell;
struct	tm	time_now;
float	cell_dif;
float	chg_dif;

httpd_handle_t iot_httpd = NULL;

#define LEDG_ON		digitalWrite( LEDG, HIGH)
#define LEDG_OFF	digitalWrite( LEDG, LOW)
#define LEDR_ON		digitalWrite( LEDR, HIGH)
#define LEDR_OFF	digitalWrite( LEDR, LOW)

float	bms_gain;
float	bms_offset;

int		charge_go;
int		charge_sel;
int		eeprom_ok;
int		wifi_ok;
int		bms_ok;
int		time_ok;
int		old_minute;

#define	MAX_CELL	3
#define	MAX_DATA	( 60 * 12)
int	cell[ MAX_CELL];
int	history[ MAX_CELL][ MAX_DATA];

float	charge_v;
float	charge_c;

int		data_num;

typedef struct	{
	char			ssid[ 32];	// ssid
	char			pawd[ 32];	// password
	unsigned char	ip[ 4];		// ip address
	unsigned char	gway[ 4];	// gateway
	unsigned int	dummy;					// dummy
	unsigned int	sum;					// check sum
} PARA_BUF;

PARA_BUF	para;

const char* ntpServer = "pool.ntp.org";
const long  gmtOffset_sec = 9*3600;
const int   daylightOffset_sec = 0;

char	response_buf[ 32768/2];

void	i2c_init( void)
{
	Wire.begin( I2C_SDA, I2C_SCL);
}

int	i2c_test( char adr)
{
	Wire.beginTransmission( adr);
	return ( Wire.endTransmission() == 0);
}

void	i2c_set1( char adr, char reg,char d)
{
	Wire.beginTransmission( adr);
	Wire.write(reg);
	Wire.write(d);
	Wire.endTransmission();
}
void	i2c_set2( char adr, char reg,char d1, char d2)
{
	Wire.beginTransmission( adr);
	Wire.write(reg);
	Wire.write(d1);
	Wire.write(d2);
	Wire.endTransmission();
}

short	i2c_get1( char adr, char reg)
{
	char	ret;

	Wire.beginTransmission( adr);
	Wire.write( reg);
	Wire.endTransmission();
	Wire.requestFrom( adr, 1);
	while( Wire.available()<1);
	ret = Wire.read() & 0x0ff;
	return ret;
}

short	i2c_get2( char adr, char reg)
{
	union	{
		char	b[ 2];
		short	s;
	} r;
	Wire.beginTransmission( adr);
	Wire.write( reg);
	Wire.endTransmission();
	Wire.requestFrom( adr, 2);
	while( Wire.available()<2);
	r.b[ 1] = Wire.read() & 0x0ff;
	r.b[ 0] = Wire.read();
	return r.s;
}

void	set_ina226( char radr, short s)
{
	union	{
		char	b[ 2];
		short	s;
	} d;
	d.s = s;
	
	i2c_set2( INA226_ADR, radr, d.b[ 1], d.b[ 0]);
}


short	get_ina226( int adr)
{
	return i2c_get2( INA226_ADR, adr);
}

float	get_voltage_ina226( void)
{
	char	adr;
	short	reg;
	float	f;

	adr = 2;
	reg = get_ina226( adr);
	f = reg;
	f = f * 0.00125;					// LSB = 1.25mV
	if( f > 999.9) f = 999.9;
	if( f < 0.0) f = 0.0;
	return f;
}

float	get_current_ina226( void)
{
	char	adr;
	short	reg;
	float	f;

	adr = 1;
	reg = get_ina226( adr);
	f = reg;
	f  = f * 2.5;						// 2.5uV INA226
	f = f / 50.0;						// mA
	return f;
}

short	get_bms16( int regadr)
{
	return i2c_get2( BMS_ADR, regadr);
}

int	get_bms8( int regadr)
{
	return i2c_get1( BMS_ADR, regadr);
}

void	set_bms8( int regadr, char d)
{
	i2c_set1( BMS_ADR, regadr, d);
}

const	float	bms_gain_table[ 32] = {
		0.365,	// 0
		0.366,	// 1
		0.367,	// 2
		0.368,	// 3
		0.369,	// 4
		0.370,	// 5
		0.371,	// 6
		0.372,	// 7
		0.373,	// 8
		0.374,	// 9
		0.375,	// 10
		0.376,	// 11
		0.377,	// 12
		0.378,	// 13
		0.379,	// 14
		0.380,	// 15
		0.381,	// 16
		0.382,	// 17
		0.383,	// 18
		0.384,	// 19
		0.385,	// 20
		0.386,	// 21
		0.387,	// 22
		0.388,	// 23
		0.389,	// 24
		0.390,	// 25
		0.391,	// 26
		0.392,	// 27
		0.393,	// 28
		0.394,	// 29
		0.395,	// 30
		0.396,	// 31

};

float	get_bms_gain( void)
{
	char	u, l, r;

	u = get_bms8( 0x50);
	l = get_bms8( 0x59);
	r = (( u & 0x0c) << 1) + (( l & 0xe0) >> 5);
	return bms_gain_table[ r];
}

float	get_bms_offset( void)
{
	char	d;

	d = get_bms8( 0x51);
	return ( float) d;
}

int	get_cell_voltage( int no)
{
	float	v;
	ushort	d;

	if( no == 0){
		d = get_bms16( 0x0c) & 0x3fff;
	} else if( no == 1){
		d = get_bms16( 0x0e) & 0x3fff;
	} else if( no == 2){
		d = get_bms16( 0x10) & 0x3fff;
	} else if( no == 3){
		d = get_bms16( 0x12) & 0x3fff;
	} else if( no == 4){
		d = get_bms16( 0x14) & 0x3fff;
	}
	v = d;
	return ( v * bms_gain + bms_offset);
}

void	save_para( void)
{
	para.sum = get_checksum( (unsigned int*)&para, sizeof( para) / 4 - 1);
	EEPROM.put<PARA_BUF>( 0, para);
	EEPROM.commit();
}

unsigned int get_checksum( unsigned int *p, int size)
{
	int	 i;
	unsigned int	sum;

	sum = 0;
	for ( i = 0 ; i < size; i++) {
		sum += *p++;
	}
	return sum;
}

void	load_para( void)
{
	unsigned int	sum;
	int						 i;

	EEPROM.get<PARA_BUF>( 0, para);
	sum = get_checksum( (unsigned int*)&para, sizeof( para) / 4 - 1);
	if(  sum != para.sum) {
		sprintf( para.ssid, "AirPort-----");
		sprintf( para.pawd, "-------------");
		para.ip[ 0] = 192;
		para.ip[ 1] = 168;
		para.ip[ 2] = 0;
		para.ip[ 3] = 101;
		para.gway[ 0] = 192;
		para.gway[ 1] = 168;
		para.gway[ 2] = 0;
		para.gway[ 3] = 1;
		save_para();
	}
}


static esp_err_t index_handler(httpd_req_t *req)
{
	int	len;
	
	len = sizeof( wifi_index_html);
     httpd_resp_send( req, wifi_index_html, len);
    return ESP_OK;
}

static esp_err_t java_handler(httpd_req_t *req)
{
	int	len;
	len = sizeof( wifi_index_js);
    httpd_resp_send( req, wifi_index_js, len);
    return ESP_OK;
}

static esp_err_t favicon_handler(httpd_req_t *req)
{
    esp_err_t res = ESP_OK;
    res = httpd_resp_send(req, (char *)wifi_favicon_ico, sizeof( wifi_favicon_ico));
    return res;
}

char*	str_int( char *p, char* name, int *d, int len)
{
	int		i;

	p += sprintf( p, "\"%s\":[", name);
	for( i = 0; i < len; i++){
		if( i < ( len-1)){
			p += sprintf( p, "%d,", d[ i]);
		} else {
			p += sprintf( p, "%d", d[ i]);
		}
	}
	p += sprintf( p, "]");
	return p;
}

void	get_data_json( char *p)
{
    *p++ = '{';

    p += sprintf(p, "\"cg\":%d", charge_go);
	p += sprintf(p, ",\"bms\":%d", bms_ok);
	p += sprintf(p, ",\"max\":%d", max_cell);
	p += sprintf(p, ",\"min\":%d", min_cell);
	p += sprintf(p, ",\"dif\":%4.0f", cell_dif);
	p += sprintf(p, ",\"chg\":%4.0f", chg_dif);
	p += sprintf(p, ",\"c0\":%d", cell[ 0]);
    p += sprintf(p, ",\"c1\":%d", cell[ 1]);
    p += sprintf(p, ",\"c2\":%d", cell[ 2]);
    p += sprintf(p, ",\"cv\":%5.3f", charge_v);
    p += sprintf(p, ",\"cc\":%5.3f", charge_c);
    p += sprintf(p, ",\"n\":%d", data_num);
    if( data_num){
	    *p++ = ',';	p = str_int( p, "h0", history[0], data_num);
	    *p++ = ',';	p = str_int( p, "h1", history[1], data_num);
	    *p++ = ',';	p = str_int( p, "h2", history[2], data_num);
    }
    *p++ = '}';
    *p++ = 0;
}

char *get_option( char *p)
{
	int	i;

	for( i = 0 ; i < 32; i++){
		if( *p == 0) return p;
		if( *p == '?') break;
		p++;
	}
	if( i == 32) return p;
	p++;
	for( ; i < 32; i++){
		if( *p == 0) return p;
		if( *p == 's') break;
		p++;
	}
	if( i == 32) return p;
	p++;
	for( ; i < 32; i++){
		if( *p == 0) return p;
		if( *p == '=') break;
		p++;
	}
	p++;
	return p;
}

static esp_err_t data_handler(httpd_req_t *req)
{
	char	b[ 32];

	int	option;
	
	option = atoi( get_option( (char*)req->uri));
	if(( charge_go == 0) && ( option == 1)){
		data_num  = 0;
	}
	charge_go = option;
	get_data_json( response_buf);
    httpd_resp_set_type(req, "application/json");
    httpd_resp_set_hdr(req, "Access-Control-Allow-Origin", "*");
    return httpd_resp_send(req, response_buf, strlen(response_buf));
}


void startServer()
{
    httpd_config_t config = HTTPD_DEFAULT_CONFIG();
    config.max_uri_handlers = 16;

    httpd_uri_t index_uri = {
        .uri = "/",
        .method = HTTP_GET,
        .handler = index_handler,
        .user_ctx = NULL};
    httpd_uri_t java_uri = {
        .uri = "/index.js",
        .method = HTTP_GET,
        .handler = java_handler,
        .user_ctx = NULL};
    httpd_uri_t favicon_uri = {
        .uri = "/favicon.ico",
        .method = HTTP_GET,
        .handler = favicon_handler,
        .user_ctx = NULL};
    httpd_uri_t data_uri = {
        .uri = "/data.json",
        .method = HTTP_GET,
        .handler = data_handler,
        .user_ctx = NULL};
        

    if( httpd_start( &iot_httpd, &config) == ESP_OK){
        httpd_register_uri_handler( iot_httpd, &index_uri);
        httpd_register_uri_handler( iot_httpd, &java_uri);
        httpd_register_uri_handler( iot_httpd, &favicon_uri);
        httpd_register_uri_handler( iot_httpd, &data_uri);
        Serial.println( "Start Server");
	}
}


void	relay_control( int onoff, int sel)
{
	switch( sel){
	case 0:	digitalWrite( RLY_SEL1, LOW);	digitalWrite( RLY_SEL2, LOW);	digitalWrite( RLY_SEL3, LOW); break;
	case 1:	digitalWrite( RLY_SEL1, HIGH);	digitalWrite( RLY_SEL2, LOW);	digitalWrite( RLY_SEL3, LOW); break;
	case 2:	digitalWrite( RLY_SEL1, LOW);	digitalWrite( RLY_SEL2, HIGH);	digitalWrite( RLY_SEL3, LOW); break;
	case 3:	digitalWrite( RLY_SEL1, LOW);	digitalWrite( RLY_SEL2, LOW);	digitalWrite( RLY_SEL3, HIGH); break;
	default:digitalWrite( RLY_SEL1, LOW);	digitalWrite( RLY_SEL2, LOW);	digitalWrite( RLY_SEL3, LOW); break;
	}
	if( onoff){
		digitalWrite( RLY_ON, HIGH);
	} else {
		digitalWrite( RLY_ON, LOW);
	}
}

int	init_wifi( void)
{
	int	i;
	int	ret = 0;
	
	WiFi.config( para.ip, para.gway, subnet, para.gway);
	WiFi.begin( para.ssid, para.pawd);
	Serial.print( para.ssid);
	for( i = 0 ; i < 10; i++) {
		if( WiFi.status() != WL_CONNECTED) {
			delay(500);
			Serial.print( ".");
		} else {
			ret = 1;
			Serial.println( " CONNECTED");

			return ret;
		}
	}
	return ret;
}

int	get_max_cell( void)
{
	int	sel;
	
	if( cell[0] > cell[ 1]){
		if( cell[ 0] > cell[ 2]){
			sel = 0;
		} else {
			sel	= 2;
		}
	} else {
		if( cell[ 1] > cell[ 2]){
			sel = 1;
		} else {
			sel	= 2;
		}
	}
	return sel;
}

int	get_min_cell( void)
{
	int	sel;
	
	if( cell[0] < cell[ 1]){
		if( cell[ 0] < cell[ 2]){
			sel = 0;
		} else {
			sel	= 2;
		}
	} else {
		if( cell[ 1] < cell[ 2]){
			sel = 1;
		} else {
			sel	= 2;
		}
	}
	return sel;
}

float	get_cell_dif( void)
{
	if( cell[ max_cell] < 4200){							// max 4200mV
		return ( cell[ max_cell] - cell[ min_cell]);
	} else {
		return ( 4200 - cell[ min_cell]);
	}
}

#define	CELL_DIF_LIMIT	10

int	do_charge( int seq)
{
	switch( seq){
	case 0:										// select cell
			charge_sel = min_cell;
			if( cell_dif < CELL_DIF_LIMIT){
				seq = 3;
			} else {
				seq++;
			}
			break;
	case 1:										// select cell
			relay_control( 1, charge_sel);		// relay on
			seq++;
			break;
	case 2:
			chg_dif = cell[ max_cell] - cell[ charge_sel];
			if( chg_dif < CELL_DIF_LIMIT){
				relay_control( 0, charge_sel);	// relay off
				seq++;
			}			
			break;
	case 3:								// charge end or next
			if( cell_dif < CELL_DIF_LIMIT){
				seq = 0;
				charge_go = 0;
			} else {
				seq = 0;
			}
			break;
	}
	return seq;
}

void	init_bms( void)
{
	int		i;
	char	b[ 64];
	float	g;
	
	Serial.println( "BMS ready");
	for( i = 0 ; i < 10; i++){
		g = get_bms_gain();
		if( g > bms_gain) bms_gain = g;
	}
	bms_offset	= get_bms_offset();
	sprintf( b, "BMS %4.3f %d", bms_gain, bms_offset);
	Serial.println( b);

	cell[ 0]	= get_cell_voltage( 0);
	cell[ 1]	= get_cell_voltage( 1);
	cell[ 2]	= get_cell_voltage( 2);
	sprintf( b, "Cell %6d %6d %6d", cell[ 2], cell[ 1], cell[ 0]);
	Serial.println( b);
}

int	printLocalTime( tm *time)
{
	if( getLocalTime( time)){
		Serial.println( time, "%A, %B %d %Y %H:%M:%S");
		return 1;
	} else {
	    Serial.println( "Failed to obtain time");
	    return 0;
	}
}

void setup() 
{
	int		i;
	float	g;
	char	b[ 64];
	
	Serial.begin(115200);
	Serial.setDebugOutput(true);
	Serial.println();

	charge_go	= 0;
	charge_seq	= 0;
	charge_sel	= 0;
	data_num	= 0;
	
	eeprom_ok = EEPROM.begin( EEP_SIZE);
	load_para();

	
	pinMode( LEDG,		OUTPUT);
	pinMode( LEDR,		OUTPUT);
	pinMode( RLY_ON,	OUTPUT);
	pinMode( RLY_SEL1,	OUTPUT);
	pinMode( RLY_SEL2,	OUTPUT);
	pinMode( RLY_SEL3,	OUTPUT);

	relay_control( 0, 0);
	LEDG_ON;
	LEDR_OFF;
	wifi_ok	= init_wifi();
	if( wifi_ok){
		startServer();
		configTime( gmtOffset_sec, daylightOffset_sec, ntpServer);
		time_ok = printLocalTime( &time_now);
	}
	i2c_init();
	set_ina226( 0, 0x4127);					// INA226  0x399f( INA219)
	charge_c	= get_current_ina226();
	charge_v	= get_voltage_ina226();
	sprintf( b, "INA226 %4.2f %4.2f", charge_c, charge_v);
	Serial.println( b);
	bms_gain = 0.0;
	bms_ok = i2c_test( BMS_ADR);
	if( bms_ok){
		init_bms();
	}
}

void loop() {
	delay( 500);
	charge_c	= get_current_ina226();
	charge_v	= get_voltage_ina226();
	if( bms_ok){
		cell[ 0]	= get_cell_voltage( 0);
		cell[ 1]	= get_cell_voltage( 1);
		cell[ 2]	= get_cell_voltage( 2);
		min_cell 	= get_min_cell();
		max_cell 	= get_max_cell();
		cell_dif 	= get_cell_dif();
		if( charge_go){
			charge_seq = do_charge( charge_seq);
		} else {
			charge_seq	= 0;
			relay_control(0, 0);
		}
	} else {
		bms_ok = i2c_test( BMS_ADR);	// check BQ7692006 
		if( bms_ok){
			init_bms();
		}
	}
//	if( time_ok){
		printLocalTime( &time_now);
		if( time_now.tm_min != old_minute){
			old_minute = time_now.tm_min;
			Serial.println( "minute passed");
			if( charge_go){
				if( data_num < MAX_DATA){
					history[ 0][ data_num] = cell[ 0];
					history[ 1][ data_num] = cell[ 1];
					history[ 2][ data_num] = cell[ 2];
					data_num++;
					Serial.println( "data saved");
				}
			}
		}
//	}
}
